@@ -1,5 +1,6 @@
 # -*- coding: utf-8 -*-
 from __future__ import absolute_import
+import ast
 import collections
 import concurrent.futures
 import contextlib
@@ -15,10 +16,10 @@ import warnings
 import numbers
 import keyword
 
+import dask.utils
 import numpy as np
 import pyarrow as pa
 import progressbar
-import psutil
 import six
 import yaml
 
@@ -35,6 +36,9 @@ osname = dict(darwin="osx", linux="linux", windows="windows")[platform.system().
 devmode = os.environ.get('VAEX_DEV', '0') == '1'
 
 
+# so that vaex can be imported without compiling the extenstions (like in readthedocs)
+has_c_extension = os.environ.get('VAEX_NO_C_EXTENSIONS', None) != "1"
+
 class AttrDict(dict):
     def __init__(self, *args, **kwargs):
         dict.__init__(self, *args, **kwargs)
@@ -272,24 +276,29 @@ def _progressbar_progressbar2(type=None, name="processing", max_value=1):
     # FormatLabel('Processed: %(value)d lines (in: %(elapsed)s)')
 
 
-def _progressbar_vaex(type=None, name="processing", max_value=1):
+def _progressbar_vaex(type=None, title="processing", max_value=1):
     import vaex.misc.progressbar as pb
-    return pb.ProgressBar(0, 1)
+    return pb.ProgressBar(0, 1, title=title)
 
-def _progressbar_widget(type=None, name="processing", max_value=1):
+def _progressbar_widget(type=None, title="processing", max_value=1):
     import vaex.misc.progressbar as pb
-    return pb.ProgressBarWidget(0, 1, name=name)
+    return pb.ProgressBarWidget(0, 1, title=title)
+
+def _progressbar_rich(type=None, title="processing", max_value=1):
+    import vaex.misc.progressbar as pb
+    return pb.ProgressBarRich(0, 1, title=title)
 
 
 _progressbar_typemap = {}
 _progressbar_typemap['progressbar2'] = _progressbar_progressbar2
 _progressbar_typemap['vaex'] = _progressbar_vaex
 _progressbar_typemap['widget'] = _progressbar_widget
+_progressbar_typemap['rich'] = _progressbar_rich
 
 
 def progressbar(type_name=None, title="processing", max_value=1):
-    type_name = type_name or 'vaex'
-    return _progressbar_typemap[type_name](name=title)
+    type_name = type_name or _progressbar_type_default
+    return _progressbar_typemap[type_name](title=title)
 
 
 def progressbar_widget():
@@ -336,15 +345,18 @@ class _progressbar_wrapper_sum(_progressbar):
 
     def add(self, name=None):
         pb = _progressbar_wrapper_sum(parent=self, name=name)
+        if self.bar and hasattr(self.bar, 'add_child'):
+            pb.bar = self.bar.add_child(pb, None, name)
         self.children.append(pb)
+        self.finished = False
+        self.fraction = sum([c.fraction for c in self.children]) / len(self.children)
+        self(self.fraction)
         return pb
 
     def add_task(self, task, name=None):
         pb = self.add(name)
         pb.oncancel = task.cancel
         task.signal_progress.connect(pb)
-        if self.bar and hasattr(self.bar, 'add_child'):
-            self.bar.add_child(pb, task, name)
 
     def __call__(self, fraction):
         if self.cancelled:
@@ -364,8 +376,8 @@ class _progressbar_wrapper_sum(_progressbar):
             elif fraction != 1:
                 if self.bar:
                     self.bar.update(fraction)
-            if self.next:
-                result = self.next(fraction)
+        if self.next:
+            result = self.next(fraction)
         if self.parent:
             assert self in self.parent.children
             result = self.parent(None) in [None, True] and result  # fraction is not used anyway..
@@ -378,9 +390,12 @@ class _progressbar_wrapper_sum(_progressbar):
         pass
 
 
-def progressbars(f=True, next=None, name=None):
+def progressbars(f=True, next=None, name=None, title=None):
     if isinstance(f, _progressbar_wrapper_sum):
-        return f
+        if title is None:
+            return f
+        else:
+            return f.add(title)
     if callable(f):
         next = f
         f = False
@@ -388,9 +403,9 @@ def progressbars(f=True, next=None, name=None):
         return _progressbar_wrapper_sum(next=next, name=name)
     else:
         if f is True:
-            return _progressbar_wrapper_sum(bar=progressbar(), next=next, name=name)
+            return _progressbar_wrapper_sum(bar=progressbar(title=title), next=next, name=name)
         elif isinstance(f, six.string_types):
-            return _progressbar_wrapper_sum(bar=progressbar(f), next=next, name=name)
+            return _progressbar_wrapper_sum(bar=progressbar(f, title=title), next=next, name=name)
         else:
             return _progressbar_wrapper_sum(next=next, name=name)
 
@@ -412,14 +427,14 @@ def confirm_on_console(topic, msg):
 
 
 def yaml_dump(f, data):
-    yaml.safe_dump(data, f, default_flow_style=False, encoding='utf-8', allow_unicode=True)
+    yaml.safe_dump(data, f, default_flow_style=False, encoding='utf-8', allow_unicode=True, sort_keys=False)
 
 
 def yaml_load(f):
     return yaml.safe_load(f)
 
 
-def write_json_or_yaml(file, data, fs_options={}, fs=None):
+def write_json_or_yaml(file, data, fs_options={}, fs=None, old_style=True):
     file, path = vaex.file.file_and_path(file, mode='w', fs_options=fs_options, fs=fs)
     try:
         if path:
@@ -427,7 +442,7 @@ def write_json_or_yaml(file, data, fs_options={}, fs=None):
         else:
             ext = '.json'  # default
         if ext == ".json":
-            json.dump(data, file, indent=2, cls=VaexJsonEncoder)
+            json.dump(data, file, indent=2, cls=VaexJsonEncoder if old_style else None)
         elif ext == ".yaml":
             yaml_dump(file, data)
         else:
@@ -436,7 +451,7 @@ def write_json_or_yaml(file, data, fs_options={}, fs=None):
         file.close()
 
 
-def read_json_or_yaml(file, fs_options={}, fs=None):
+def read_json_or_yaml(file, fs_options={}, fs=None, old_style=True):
     file, path = vaex.file.file_and_path(file, fs_options=fs_options, fs=fs)
     try:
         if path:
@@ -444,7 +459,7 @@ def read_json_or_yaml(file, fs_options={}, fs=None):
         else:
             ext = '.json'  # default
         if ext == ".json":
-            return json.load(file, cls=VaexJsonDecoder) or {}
+            return json.load(file, cls=VaexJsonDecoder if old_style else None) or {}
         elif ext == ".yaml":
             return yaml_load(file) or {}
         else:
@@ -453,24 +468,8 @@ def read_json_or_yaml(file, fs_options={}, fs=None):
         file.close()
 
 
-# from http://stackoverflow.com/questions/5121931/in-python-how-can-you-load-yaml-mappings-as-ordereddicts
-
-_mapping_tag = yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG
-
-
-def dict_representer(dumper, data):
-    return dumper.represent_dict(data.iteritems() if hasattr(data, "iteritems") else data.items())
-
-
-def dict_constructor(loader, node):
-    return collections.OrderedDict(loader.construct_pairs(node))
-
-
-yaml.add_representer(collections.OrderedDict, dict_representer, yaml.SafeDumper)
-yaml.add_constructor(_mapping_tag, dict_constructor, yaml.SafeLoader)
-
-
 def check_memory_usage(bytes_needed, confirm):
+    psutil = optional_import('psutil')
     if bytes_needed > psutil.virtual_memory().available:
         if bytes_needed < (psutil.virtual_memory().available + psutil.swap_memory().free):
             text = "Action requires %s, you have enough swap memory available but it will make your computer slower, do you want to continue?" % (
@@ -557,10 +556,11 @@ def valid_identifier(name):
     return name.isidentifier() and not keyword.iskeyword(name)
 
 
-def find_valid_name(name, used=[]):
+def find_valid_name(name, used=None):
+    if used is None:
+        used = []
     if isinstance(name, int):
         name = str(name)
-    first, rest = name[0], name[1:]
     if name in used:
         nr = 1
         while name + ("_%d" % nr) in used:
@@ -569,15 +569,17 @@ def find_valid_name(name, used=[]):
     return name
 
 
-def _python_save_name(name, used=[]):
-	first, rest = name[0], name[1:]
-	name = re.sub("[^a-zA-Z_]", "_", first) +  re.sub("[^a-zA-Z_0-9]", "_", rest)
-	if name in used:
-		nr = 1
-		while name + ("_%d" % nr) in used:
-			nr += 1
-		name = name + ("_%d" % nr)
-	return name
+def _python_save_name(name, used=None):
+    if used is None:
+        used = []
+    first, rest = name[0], name[1:]
+    name = re.sub("[^a-zA-Z_]", "_", first) + re.sub("[^a-zA-Z_0-9]", "_", rest)
+    if name in used:
+        nr = 1
+        while name + ("_%d" % nr) in used:
+            nr += 1
+        name = name + ("_%d" % nr)
+    return name
 
 
 @contextlib.contextmanager
@@ -671,6 +673,13 @@ def _normalize_selection_name(name):
         return name
 
 
+def _normalize_selection(selection):
+    if isinstance(selection, (list, tuple)):
+        return type(selection)([_normalize_selection_name(k) for k in selection])
+    else:
+        return _normalize_selection_name(selection)
+
+
 def _parse_n(n):
     if isinstance(n, six.string_types):
         if n == "normalize":
@@ -826,15 +835,15 @@ def as_contiguous(ar):
 
 
 def _split_and_combine_mask(arrays):
-	'''Combines all masks from a list of arrays, and logically ors them into a single mask'''
-	masks = [np.ma.getmaskarray(block) for block in arrays if np.ma.isMaskedArray(block)]
-	arrays = [block.data if np.ma.isMaskedArray(block) else block for block in arrays]
-	mask = None
-	if masks:
-		mask = masks[0].copy()
-		for other in masks[1:]:
-			mask |= other
-	return arrays, mask
+    '''Combines all masks from a list of arrays, and logically ors them into a single mask'''
+    masks = [np.ma.getmaskarray(block) for block in arrays if np.ma.isMaskedArray(block)]
+    arrays = [block.data if np.ma.isMaskedArray(block) else block for block in arrays]
+    mask = None
+    if masks:
+        mask = masks[0].copy()
+        for other in masks[1:]:
+            mask |= other
+    return arrays, mask
 
 def gen_to_list(fn=None, wrapper=list):
     '''A decorator which wraps a function's return value in ``list(...)``.
@@ -886,7 +895,7 @@ def find_type_from_dtype(namespace, prefix, dtype, transient=True, support_non_n
         if postfix == '>f8':
             postfix = 'float64'
         if dtype.kind == "M":
-            postfix = "uint64"
+            postfix = "int64"
         if dtype.kind == "m":
             postfix = "int64"
         # for object there is no non-native version
@@ -1025,6 +1034,20 @@ def div_ceil(n, d):
     return (n + d - 1) // d
 
 
+def get_env_memory(key, default=None):
+    value = os.environ.get(key, default)
+    if value is not None:
+        try:
+            value = ast.literal_eval(value)
+        except:
+            pass
+        if isinstance(value, str):
+            value = dask.utils.parse_bytes(value)
+        if not isinstance(value, int):
+            raise TypeError(f"Expected env var {key} to be of integer type")
+    return value
+
+
 def get_env_type(type, key, default=None):
     '''Get an env var named key, and cast to type
 
@@ -1049,5 +1072,15 @@ def get_env_type(type, key, default=None):
         # support empty strings
         value = default
     if value is not None:
-        import ast
-        return type(ast.literal_eval(str(value)))
+        return type(ast.literal_eval(repr(value)))
+
+
+def dropnan(sequence, expect=None):
+    original_type = type(sequence)
+    sequence = list(sequence)
+    non_nan = [k for k in sequence if k == k]
+    if expect is not None:
+        assert len(sequence) - len(non_nan) == 1, "expected 1 nan value"
+    return original_type(non_nan)
+
+_progressbar_type_default = get_env_type(str, 'VAEX_PROGRESS_TYPE', 'vaex')
